/*
 * Copyright (C) 2010 SENJU Jiro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _GNU_SOURCE

#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>

#include <libaio.h>

int done;
char *fname;

void libaio_cb(io_context_t ctx, struct iocb *iocb, long res, long res2)
{
	done = 1;
	printf("%s: io_submit from fd %d completed\n",
	       __func__, iocb->aio_fildes);
	printf("%s, %ld read\n", fname, res);
}

int main(int argc, char *argv[])
{
	int err, fd;
	char *p;
	io_context_t libaio;
	struct iocb *pcb, laiocb;

	fd = open(argv[1], O_RDONLY | O_DIRECT);
	assert(fd >= 0);
	err = posix_memalign((void *)&p, 512, BUFSIZ);
	assert(!err);

	fname = argv[1];
	err = io_queue_init(1, &libaio);
	assert(!err);
	io_prep_pread(&laiocb, fd, p, BUFSIZ, 0);
	io_set_callback(&laiocb, libaio_cb);

	pcb = &laiocb;
	err = io_submit(libaio, 1, &pcb);
	assert(err == 1);

	/* do other jobs */

	err = io_queue_run(libaio);
	if (!done)
		puts("AIO undone");

	return err;
}
