/*
 * Copyright (C) 2009-2010 SENJU Jiro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef __LAIO_H__
#define __LAIO_H__

#include <asm/unistd.h>
#include <sys/syscall.h>
#include <linux/aio_abi.h>
#include <aio.h> /* glibc */
#include <time.h>
#include <unistd.h>

static inline int io_setup(unsigned nr_events, aio_context_t *ctxp)
{
	return syscall(__NR_io_setup, nr_events, ctxp);
}

static inline int io_destroy(aio_context_t ctx)
{
	return syscall(__NR_io_destroy, ctx);
}

static inline int io_getevents(aio_context_t ctx, long min_nr, long nr,
			       struct io_event *events,
			       struct timespec *timeout)
{
	return syscall(__NR_io_getevents, ctx, min_nr, nr, events, timeout);
}

static inline int io_submit(aio_context_t ctx, long nr, struct iocb **iocbpp)
{
	return syscall(__NR_io_submit, ctx, nr, iocbpp);
}

static inline int io_cancel(aio_context_t ctx, struct iocb *iocb,
			    struct io_event *result)
{
	return syscall(__NR_io_cancel, ctx, iocb, result);
}

/* ---------------------------------------------------------------------- */

struct la_counter {
	pthread_mutex_t	mtx;
	int		cur, lim;
};

struct la_lio {
	struct la_counter	nio;
	struct sigevent		sev;
};

struct la_cb {
	struct aiocb	la_aiocb;
	struct iocb	la_iocb;
	pid_t		la_pid;
	struct la_lio	*la_lio;
};

#ifdef __USE_LARGEFILE64
struct la_cb64 {
	struct aiocb64	la_aiocb;
	struct iocb	la_iocb;
	pid_t		la_pid;
	struct la_lio	*la_lio;
};
#endif

#ifdef __LAIO_LIB64__
#define la_cb		la_cb64
#endif

/* ---------------------------------------------------------------------- */

#ifdef __USE_GNU
void laio_init(const struct aioinit *init);
#endif

int laio_read(struct la_cb *cb);
int laio_write(struct la_cb *cb);
int laio_fsync(int op, struct la_cb *cb);
ssize_t laio_return(struct la_cb *cb);
int laio_error(const struct la_cb *cb);
int laio_suspend(const struct la_cb * const cbpl[], int n,
		 const struct timespec *timeout);
int laio_cancel(int fd, struct la_cb *cb);
int llio_listio(int mode, struct la_cb * const cbpl[], int nent,
		struct sigevent *sig);

#ifdef __USE_LARGEFILE64
int laio_read64(struct la_cb64 *cb);
int laio_write64(struct la_cb64 *cb);
int laio_fsync64(int op, struct la_cb64 *cb);
ssize_t laio_return64(struct la_cb64 *cb);
int laio_error64(const struct la_cb64 *cb);
int laio_suspend64(const struct la_cb64 * const cbpl[], int n,
		   const struct timespec *timeout);
int laio_cancel64(int fd, struct la_cb64 *cb);
int llio_listio64(int mode, struct la_cb64 * const cbpl[], int nent,
		  struct sigevent *sig);
#endif /* __USE_LARGEFILE64 */

#ifndef __LAIO_LIB__
#define aio_fildes	la_aiocb.aio_fildes
#define aio_lio_opcode	la_aiocb.aio_lio_opcode
#define aio_reqprio	la_aiocb.aio_reqprio
#define aio_buf		la_aiocb.aio_buf
#define aio_nbytes	la_aiocb.aio_nbytes
#define aio_sigevent	la_aiocb.aio_sigevent
#define aio_offset	la_aiocb.aio_offset

#define aiocb		la_cb
#define aiocb64		la_cb64

#ifdef __USE_GNU
#define aio_init(init)			laio_init(init)
#endif

/* extention for Linux AIO */
int laio_fin(void);

#define aio_read(cb)			laio_read(cb)
#define aio_write(cb)			laio_write(cb)
#define aio_fsync(op, cb)		laio_fsync(op, cb)
#define aio_return(cb)			laio_return(cb)
#define aio_error(cb)			laio_error(cb)
#define aio_suspend(cbpl, n, to)	laio_suspend(cbpl, n, to)
#define aio_cancel(fd, cb)		laio_cancel(fd, cb)
#define lio_listio(mode, cbpl, n, sig)	llio_listio(mode, cbpl, n, sig)

#ifdef __USE_LARGEFILE64
#define aio_read64(cb)				laio_read64(cb)
#define aio_write64(cb)				laio_write64(cb)
#define aio_fsync64(op, cb)			laio_fsync64(op, cb)
#define aio_return64(cb)			laio_return64(cb)
#define aio_error64(cb)				laio_error64(cb)
#define aio_suspend64(cbpl, n, to)		laio_suspend64(cbpl, n, to)
#define aio_cancel64(fd, cb)			laio_cancel64(fd, cb)
#define lio_listio64(mode, cbpl, n, sig)	llio_listio64(mode, cbpl, n, sig)
#endif /* __USE_LARGEFILE64 */
#endif /* __LAIO_LIB__ */

#endif /* __LAIO_H__ */
